# this python script performs the same calculation as example_ad.ipynb

import json
import csv
import argparse
import os

from aivis_engine_v2_ad_sdk_python import (
    EngineLogger,
    AnomalyDetectionSetup,
    AnomalyDetectionData,
    AnomalyDetectionTraining,
    AnomalyDetectionInference,
    DtoFloatDataPoint,
)


# ---------------------------------------------------------------------------------
# Data location

TRAINING_DATA_FILE = "train_ad.csv"
INFERENCE_DATA_FILE = "eval_ad.csv"
TARGET_SIGNAL_ID = "L-1"


# Input and output folder specified via args
def from_args():
    parser = argparse.ArgumentParser(
        description="Example project for signal prediction"
    )
    # data is expected in input folder
    parser.add_argument(
        "--input", "-i", type=str, nargs=1, required=True, help="Path to input folder"
    )
    # output files will be written to output folder
    parser.add_argument(
        "--output", "-o", type=str, nargs=1, required=True, help="Path to output folder"
    )
    args = parser.parse_args()
    # create output folder if not existing
    if not os.path.exists(args.output[0]):
        os.mkdir(args.output[0])
    return {
        "input_folder": args.input[0],
        "output_folder": args.output[0],
    }


# ---------------------------------------------------------------------------------
# Logger and Threads


class Logger(EngineLogger):
    def log(self, level, thread, module, message):
        min_logging_level = 3
        if level <= min_logging_level:
            print("\t... %s" % message)


def setup_logger_and_threads():
    print("Initializing AnomalyDetectionSetup: logging and threads ...")
    AnomalyDetectionSetup.register_logger(Logger())
    max_threads = 0  # number of threads will be chosen automatically with 0
    AnomalyDetectionSetup.init_thread_count(max_threads)
    print("\t... done!")
    return


# ---------------------------------------------------------------------------------
# Reading and Writing


def read_json(filename):
    print("\t... reading file '%s' ..." % filename)
    with open(filename, "r") as file:
        return json.loads(file.read())


def write_json(filename, obj):
    print("\t... writing file '%s' ..." % filename)
    with open(filename, "w") as file:
        file.write(json.dumps(obj, indent=2))
    return


# convert [DtoFloatDataPoint] to json
def float_data_points_to_timeseries_json(list_object):
    return {
        "_type": "Float",
        "dataPoints": [
            {
                "timestamp": int(point_object.timestamp),
                "value": float(point_object.value),
            }
            for point_object in list_object
        ],
    }


# convert [DtoFloatDataPointWithNextNormal] to json
def float_data_points_with_next_normal_to_json(list_object):
    return {
        "_type": "FloatWithNextNormal",
        "dataPoints": [
            {
                "timestamp": int(point_object.timestamp),
                "value": float(point_object.value),
                "features": [
                    {
                        "feature": int(feature.feature),
                        "value": float(feature.value),
                        "nextNormalValue": float(feature.next_normal_value),
                        "rating": float(feature.rating),
                    }
                    for feature in point_object.features
                ],
            }
            for point_object in list_object
        ],
    }


# Main data reading function
# Float signals are added to an aivis data context via their id and a list of DtoFloatDataPoint.
# (Analogously for boolean and string signals.)
# The rest of the following function can be adapted if the data comes in a different format as in this example, e.g. in parquet files.
def read_csv_convert_to_aivis_format(filename):
    print("\t... reading file '%s' ..." % filename)

    # create empty data context for training data
    data_context = AnomalyDetectionData.create()

    # get line numbers to show nice progress bar
    line_count = sum(1 for _ in open(filename))

    # read file
    with open(filename, newline="\n") as csvfile:
        # in this example, signal ids are in the header, first column is timestamps
        reader = csv.reader(csvfile, delimiter=",", quotechar='"')
        headers = next(reader)
        signals = list(map(lambda h: {"signal_id": h, "dataPoints": []}, headers[1:]))

        # show nice progress bar
        progress_bar = -1
        line = 0

        for row in reader:
            # show nice progress bar
            progess_bar_new = round(line / (line_count / 10)) * 10
            if progress_bar != progess_bar_new:
                progress_bar = progess_bar_new
                print("\r\t... processing signals: %d%% ..." % progress_bar, end="")
            line += 1

            # get timestamp of this row, which is in the csv's first column
            timestamp = int(row[0])

            # add points to all signal's list (if value exists)
            for i, column in enumerate(row[1:], start=0):
                if column:
                    # in this example, all signals are float signals
                    value = float(column)
                    signals[i]["dataPoints"].append(DtoFloatDataPoint(timestamp, value))
        print("... reading from file is done!")

        print("\t... adding signals to data context:")
        # for larger archives it would be better to read and then immediately add to data context in chunks
        # e.g. signal per signal
        for signal in signals:
            data_context.add_float_signal(signal["signal_id"], signal["dataPoints"])
            print(
                "\t   ... Float Signal "
                + signal["signal_id"]
                + " had "
                + str(len(signal["dataPoints"]))
                + " data points"
            )

        print("... data context is ready!")
        return data_context


# ---------------------------------------------------------------------------------
# MAIN


def main():
    folder_config = from_args()  # parsing args as folder_config
    setup_logger_and_threads()

    print("Loading training data ...")
    training_data = read_csv_convert_to_aivis_format(
        os.path.join(folder_config["input_folder"], TRAINING_DATA_FILE)
    )

    # print some signal to illustrate type DtoFloatDataPoint
    print("Example signal!")
    some_signal_id = "L-3"
    print("\t... Signal id:", some_signal_id)
    some_signal_data_points = training_data.get_float_signal(some_signal_id)
    print("\t... Signal length:", len(some_signal_data_points))
    print("\t... Signal entry number 42:", some_signal_data_points[42])
    print(
        "\t... Signal entry number 42 timestamp (unix/millisec):",
        some_signal_data_points[42].timestamp,
    )
    print("\t... Signal entry number 42 value:", some_signal_data_points[42].value)

    # ---- TRAINING ----

    print("Creating training ...")
    # build training config - for more options see documentation
    training_config = json.dumps(
        {
            "target": {"signal": TARGET_SIGNAL_ID},
        }
    )

    # create training and train the model
    training = AnomalyDetectionTraining.create(training_data, training_config)
    print("\t... done!")

    print("Save training files ...")
    model = training.get_model()
    write_json(
        os.path.join(folder_config["output_folder"], "model.json"), json.loads(model)
    )

    training_report = json.loads(training.get_report())  # training infos
    write_json(
        os.path.join(folder_config["output_folder"], "training-report.json"),
        training_report,
    )
    print("\t... done!")

    # ---- INFERENCE ----

    print("Loading inference data ...")
    # reading all the signals for simplicity
    # filtering out the relevant ones in the model's inference data spec would also be enough
    inference_data = read_csv_convert_to_aivis_format(
        os.path.join(folder_config["input_folder"], INFERENCE_DATA_FILE)
    )

    print("Creating inference ...")
    # build inference config
    inference_config = json.dumps({"skipOnInsufficientData": True})
    # create inference from training object
    # via model is also an option: AnomalyDetectionInference.create_by_model(model_json, inference_config)
    inference = AnomalyDetectionInference.create_by_training(training, inference_config)
    print("\t... done!")

    print("Calculating scores ...")
    # choose inference timestamps: selecting timestamps from evaluation data,
    # but of course any timestamp where the inference data meets the model's inference data specification is possible
    timestamps = list(range(1465632000000, 1466786640000, 60000))
    # infer anomaly scores
    scores = inference.infer(inference_data, timestamps)
    print("\t... done!")
    print("Number of scores: ", len(scores))

    """ 
    # comment in you want to calculate next normal point
    print("Calculating scores with next normal point ...")
    # choose inference timestamps: selecting timestamps from evaluation data, here we choose a coarser granularity than above
    # because calculation time of endpoint infer_with_next_normal is much larger
    timestamps_next_normal = list(range(1465632000000, 1466786640000, 600000))
    # infer anomaly scores with next normal point 
    scores_with_next_normal = inference.infer_with_next_normal(inference_data, timestamps_next_normal, 0.5)
    print("\t... done!")
    print("Number of scores with next normal: ", len(scores_with_next_normal))
    """

    print("Save inference file ...")
    write_json(
        os.path.join(folder_config["output_folder"], "scores.json"),
        float_data_points_to_timeseries_json(scores),
    )
    # comment in if scoresWithNextNormal were computed to save as json
    # write_json(os.path.join(folder_config["output_folder"], "scores_with_next_normal.json"), float_data_points_with_next_normal_to_json(scores_with_next_normal))
    print("\t... done!")

    print("Destroy everything ...")
    # destroying can be done once the respective object is not needed anymore - do this if memory is an issue
    training_data.destroy()
    training.destroy()
    inference_data.destroy()
    inference.destroy()
    print("\t... done!")

    # ---- Toolbox HTML Report ----
    from aivis_engine_v2_toolbox.api import build_anomaly_detection_report

    build_anomaly_detection_report(
        {"title": "Aivis Engine V2 - Anomaly Detection Example: Condition Monitoring",
         "scores": {"file": os.path.join(folder_config["output_folder"], "scores.json")},
         "trainingReportFile": os.path.join(folder_config["output_folder"], "training-report.json"),
         "outputFile": os.path.join(folder_config["output_folder"], "aivis-engine-v2-anomaly-detection-example.html")})

    print("ALL DONE!")
    return


# run main function
main()
